@@ -14,19 +14,20 @@ from abc import ABCMeta
 import locale
 from collections.abc import MutableSequence
 from urllib.parse import urlparse
-from typing import cast, Any, Callable, ClassVar, Dict, FrozenSet, List, \
+from typing import cast, Any, Callable, ClassVar, Dict, List, \
     MutableMapping, Optional, Tuple, Type, Union
 
-from ..helpers import normalize_sequence_type
+from ..helpers import upper_camel_case, is_ncname, ordinal
 from ..exceptions import ElementPathError, ElementPathTypeError, \
     ElementPathValueError, MissingContextError, xpath_error
 from ..namespaces import NamespacesType, XSD_NAMESPACE, XML_NAMESPACE, \
-    XLINK_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, XQT_ERRORS_NAMESPACE, \
+    XPATH_FUNCTIONS_NAMESPACE, XQT_ERRORS_NAMESPACE, \
     XSD_NOTATION, XSD_ANY_ATOMIC_TYPE, get_prefixed_name
+from ..collations import UNICODE_COLLATION_BASE_URI, UNICODE_CODEPOINT_COLLATION
 from ..datatypes import UntypedAtomic, AtomicValueType, QName
-from ..xpath_token import UNICODE_CODEPOINT_COLLATION, XPathToken, \
-    XPathFunction, XPathConstructor
+from ..xpath_tokens import NargsType, XPathToken, ProxyToken, XPathFunction, XPathConstructor
 from ..xpath_context import XPathContext
+from ..sequence_types import is_sequence_type, match_sequence_type
 from ..schema_proxy import AbstractSchemaProxy
 from ..xpath1 import XPath1Parser
 
@@ -71,111 +72,11 @@ class XPath2Parser(XPath1Parser):
     Default is 'node()*'.
     """
     version = '2.0'
-
-    SYMBOLS: ClassVar[FrozenSet[str]] = XPath1Parser.SYMBOLS | {
-        'union', 'intersect', 'instance', 'castable', 'if', 'then', 'else', 'for', 'to',
-        'some', 'every', 'in', 'satisfies', 'item', 'satisfies', 'cast', 'treat',
-        'return', 'except', '?', 'as', 'of',
-
-        # Comments
-        '(:', ':)',
-
-        # Value comparison operators
-        'eq', 'ne', 'lt', 'le', 'gt', 'ge',
-
-        # Node comparison operators
-        'is', '<<', '>>',
-
-        # Mathematical operators
-        'idiv',
-
-        # Node type functions
-        'document-node', 'schema-attribute', 'element', 'schema-element',
-        'attribute', 'empty-sequence',
-
-        # Accessor functions
-        'node-name', 'nilled', 'data', 'base-uri', 'document-uri',
-
-        # Number functions
-        'abs', 'round-half-to-even',
-
-        # Aggregate functions
-        'avg', 'min', 'max',
-
-        # String functions
-        'codepoints-to-string', 'string-to-codepoints', 'compare', 'codepoint-equal',
-        'string-join', 'normalize-unicode', 'upper-case', 'lower-case', 'encode-for-uri',
-        'iri-to-uri', 'escape-html-uri', 'ends-with',
-
-        # General functions for sequences
-        'distinct-values', 'empty', 'exists', 'index-of', 'insert-before', 'remove',
-        'reverse', 'subsequence', 'unordered',
-
-        # Cardinality functions for sequences
-        'zero-or-one', 'one-or-more', 'exactly-one',
-
-        # Comparing function for sequences
-        'deep-equal',
-
-        # Pattern matching functions
-        'matches', 'replace', 'tokenize',
-
-        # Functions on anyURI
-        'resolve-uri',
-
-        # Functions for extracting fragments from xs:duration
-        'years-from-duration', 'months-from-duration', 'days-from-duration',
-        'hours-from-duration', 'minutes-from-duration', 'seconds-from-duration',
-
-        # Functions for extracting fragments from xs:dateTime
-        'year-from-dateTime', 'month-from-dateTime', 'day-from-dateTime', 'hours-from-dateTime',
-        'minutes-from-dateTime', 'seconds-from-dateTime', 'timezone-from-dateTime',
-
-        # Functions for extracting fragments from xs:date
-        'year-from-date', 'month-from-date', 'day-from-date', 'timezone-from-date',
-
-        # Functions for extracting fragments from xs:time
-        'hours-from-time', 'minutes-from-time', 'seconds-from-time', 'timezone-from-time',
-
-        # Timezone adjustment functions
-        'adjust-dateTime-to-timezone', 'adjust-date-to-timezone', 'adjust-time-to-timezone',
-
-        # Functions Related to QNames (QName function is also a constructor)
-        'QName', 'local-name-from-QName', 'prefix-from-QName', 'local-name-from-QName',
-        'namespace-uri-from-QName', 'namespace-uri-for-prefix', 'in-scope-prefixes',
-        'resolve-QName',
-
-        # Static context functions
-        'default-collation', 'static-base-uri',
-
-        # Dynamic context functions
-        'current-dateTime', 'current-date', 'current-time', 'implicit-timezone',
-
-        # Node set functions
-        'root',
-
-        # Error function and trace function
-        'error', 'trace',
-
-        # XSD builtins constructors ('string', 'boolean' and 'QName' are
-        # already registered as functions)
-        'normalizedString', 'token', 'language', 'Name', 'NCName', 'ENTITY', 'ID',
-        'IDREF', 'NMTOKEN', 'anyURI', 'NOTATION', 'decimal', 'int', 'integer', 'long',
-        'short', 'byte', 'double', 'float', 'nonNegativeInteger', 'positiveInteger',
-        'nonPositiveInteger', 'negativeInteger', 'unsignedLong', 'unsignedInt',
-        'unsignedShort', 'unsignedByte', 'dateTime', 'date', 'time', 'gDay', 'gMonth',
-        'gYear', 'gMonthDay', 'gYearMonth', 'duration', 'dayTimeDuration',
-        'yearMonthDuration', 'dateTimeStamp', 'base64Binary', 'hexBinary', 'untypedAtomic',
-
-        # Functions and Operators that Generate Sequences ('id' changes but
-        # is already registered)
-        'element-with-id', 'idref', 'doc', 'doc-available', 'collection',
-    }
+    default_collation = UNICODE_CODEPOINT_COLLATION
 
     DEFAULT_NAMESPACES: ClassVar[Dict[str, str]] = {
         'xml': XML_NAMESPACE,
         'xs': XSD_NAMESPACE,
-        'xlink': XLINK_NAMESPACE,
         'fn': XPATH_FUNCTIONS_NAMESPACE,
         'err': XQT_ERRORS_NAMESPACE
     }
@@ -185,13 +86,23 @@ class XPath2Parser(XPath1Parser):
         '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '(', '{'
     }
 
+    # https://www.w3.org/TR/xpath20/#id-reserved-fn-names
+    RESERVED_FUNCTION_NAMES = {
+        'attribute', 'comment', 'document-node', 'element', 'empty-sequence',
+        'if', 'item', 'node', 'processing-instruction', 'schema-attribute',
+        'schema-element', 'text', 'typeswitch',
+    }
+
     function_signatures: Dict[Tuple[QName, int], str] = XPath1Parser.function_signatures.copy()
     namespaces: Dict[str, str]
     token: XPathToken
     next_token: XPathToken
 
+    @staticmethod
+    def tracer(trace_data: str) -> None:
+        """Trace data collector"""
+
     def __init__(self, namespaces: Optional[NamespacesType] = None,
-                 variable_types: Optional[Dict[str, str]] = None,
                  strict: bool = True,
                  compatibility_mode: bool = False,
                  default_collation: Optional[str] = None,
@@ -200,17 +111,32 @@ class XPath2Parser(XPath1Parser):
                  xsd_version: Optional[str] = None,
                  schema: Optional[AbstractSchemaProxy] = None,
                  base_uri: Optional[str] = None,
+                 variable_types: Optional[Dict[str, str]] = None,
                  document_types: Optional[Dict[str, str]] = None,
                  collection_types: Optional[Dict[str, str]] = None,
                  default_collection_type: str = 'node()*') -> None:
 
         super(XPath2Parser, self).__init__(namespaces, strict)
-        self._compatibility_mode = compatibility_mode
-        self._default_collation = default_collation
+        self.compatibility_mode = compatibility_mode
+
+        if default_collation is not None:
+            self.default_collation = default_collation
+        else:
+            # Obtain the current collation locale using setlocale() with `None`.
+            # Consider only configured UTF-8 encodings, otherwise keep Unicode
+            # Codepoint Collation.
+            _locale = locale.setlocale(locale.LC_COLLATE, None)
+            if '.' in _locale:
+                language_code, encoding = _locale.split('.')
+                if encoding.lower() == 'utf-8':
+                    self.default_collation = f'{UNICODE_COLLATION_BASE_URI}?lang={language_code}'
+
         self._xsd_version = xsd_version if xsd_version is not None else '1.0'
 
         if default_namespace is not None:
-            self.namespaces[''] = default_namespace
+            self.default_namespace = self.namespaces[''] = default_namespace
+        else:
+            self.default_namespace = self.namespaces.get('', '')
 
         if function_namespace is not None:
             self.function_namespace = function_namespace
@@ -225,66 +151,64 @@ class XPath2Parser(XPath1Parser):
 
         if not variable_types:
             self.variable_types = {}
-        elif all(self.is_sequence_type(v) for v in variable_types.values()):
-            self.variable_types = {
-                k: normalize_sequence_type(v) for k, v in variable_types.items()
-            }
+        elif all(is_sequence_type(v, self) for v in variable_types.values()):
+            self.variable_types = variable_types.copy()
         else:
             raise ElementPathValueError('invalid sequence type for in-scope variable types')
 
         self.base_uri = None if base_uri is None else urlparse(base_uri).geturl()
 
         if document_types:
-            if any(not self.is_sequence_type(v) for v in document_types.values()):
+            if any(not is_sequence_type(v, self) for v in document_types.values()):
                 raise ElementPathValueError('invalid sequence type in document_types argument')
         self.document_types = document_types
 
         if collection_types:
-            if any(not self.is_sequence_type(v) for v in collection_types.values()):
+            if any(not is_sequence_type(v, self) for v in collection_types.values()):
                 raise ElementPathValueError('invalid sequence type in collection_types argument')
         self.collection_types = collection_types
 
-        if not self.is_sequence_type(default_collection_type):
+        if not is_sequence_type(default_collection_type, self):
             raise ElementPathValueError('invalid sequence type for '
                                         'default_collection_type argument')
         self.default_collection_type = default_collection_type
 
+    def __repr__(self) -> str:
+        args = []
+        if self.compatibility_mode:
+            args.append('compatibility_mode=True')
+        if self.default_collation != UNICODE_CODEPOINT_COLLATION:
+            args.append(f'default_collation={self.default_collation!r}')
+        if self.function_namespace != XPATH_FUNCTIONS_NAMESPACE:
+            args.append(f'function_namespace={self.function_namespace!r}')
+        if self._xsd_version != '1.0':
+            args.append(f'xsd_version={self._xsd_version!r}')
+        if self.schema is not None:
+            args.append(f'schema={self.schema!r}')
+        if self.base_uri is not None:
+            args.append(f'base_uri={self.base_uri!r}')
+        if self.variable_types:
+            args.append(f'variable_types={self.variable_types!r}')
+        if self.document_types:
+            args.append(f'document_types={self.document_types!r}')
+        if self.collection_types:
+            args.append(f'collection_types={self.collection_types!r}')
+        if self.default_collection_type != 'node()*':
+            args.append(f'default_collection_type={self.default_collection_type!r}')
+        if not args:
+            return super().__repr__()
+
+        repr_string = super().__repr__()[:-1]
+        if repr_string.endswith('('):
+            return f"{repr_string}{', '.join(args)})"
+        return f"{repr_string}, {', '.join(args)})"
+
     def __getstate__(self) -> Dict[str, Any]:
         state = self.__dict__.copy()
         state.pop('symbol_table', None)
         state.pop('tokenizer', None)
         return state
 
-    @property
-    def compatibility_mode(self) -> bool:
-        return self._compatibility_mode
-
-    @compatibility_mode.setter
-    def compatibility_mode(self, value: bool) -> None:
-        self._compatibility_mode = value
-
-    @property
-    def default_collation(self) -> str:
-        if self._default_collation is not None:
-            return self._default_collation
-
-        language_code, encoding = locale.getdefaultlocale()
-
-        if language_code is None:
-            return UNICODE_CODEPOINT_COLLATION
-        elif encoding is None or not encoding:
-            return language_code
-        else:
-            collation = f'{language_code}.{encoding}'
-            if collation != 'en_US.UTF-8':
-                return collation
-            else:
-                return UNICODE_CODEPOINT_COLLATION
-
-    @property
-    def default_namespace(self) -> Optional[str]:
-        return self.namespaces.get('')
-
     @property
     def xsd_version(self) -> str:
         if self.schema is None:
@@ -295,17 +219,17 @@ class XPath2Parser(XPath1Parser):
         except (AttributeError, NotImplementedError):
             return self._xsd_version
 
-    def advance(self, *symbols: str) -> XPathToken:
-        super(XPath2Parser, self).advance(*symbols)
+    def advance(self, *symbols: str,  message: Optional[str] = None) -> XPathToken:
+        super(XPath2Parser, self).advance(*symbols, message=message)
 
         if self.next_token.symbol == '(:':
-            try:
-                self.token.unexpected(':')
-            except AttributeError:
-                pass
+            # Parses and consumes an XPath 2.0 comment. A comment is delimited
+            # by symbols '(:' and ':)' and can be nested. The current token is
+            # saved and restored after parsing the entire comment. Comments
+            # cannot be inside a prefixed name ':' specification.
+            self.token.unexpected(':')
+            token = self.token
 
-            # Parses and consumes an XPath 2.0 comment. A comment is
-            # delimited by symbols '(:' and ':)' and can be nested.
             comment_level = 1
             while comment_level:
                 self.advance_until('(:', ':)')
@@ -314,12 +238,14 @@ class XPath2Parser(XPath1Parser):
                 else:
                     comment_level += 1
             self.advance(':)')
-            self.next_token.unexpected(':')
 
-        return self.next_token
+            self.next_token.unexpected(':')
+            self.token = token
+
+        return self.token
 
     @classmethod
-    def constructor(cls, symbol: str, bp: int = 0, nargs: int = 1,
+    def constructor(cls, symbol: str, bp: int = 90, nargs: NargsType = 1,
                     sequence_types: Union[Tuple[()], Tuple[str, ...], List[str]] = (),
                     label: Union[str, Tuple[str, ...]] = 'constructor function') \
             -> Callable[[Callable[..., Any]], Callable[..., Any]]:
@@ -329,18 +255,27 @@ class XPath2Parser(XPath1Parser):
                 self.parser.advance('(')
                 self[0:] = self.parser.expression(5),
                 if self.parser.next_token.symbol == ',':
-                    raise self.wrong_nargs('Too many arguments: expected at most 1 argument')
+                    msg = 'Too many arguments: expected at most 1 argument'
+                    raise self.error('XPST0017', msg)
                 self.parser.advance(')')
                 self.value = None
             except SyntaxError:
                 raise self.error('XPST0017') from None
+
+            if self[0].symbol == '?':
+                self.to_partial_function()
             return self
 
         def evaluate_(self: XPathConstructor, context: Optional[XPathContext] = None) \
                 -> Union[List[None], AtomicValueType]:
+            if self.context is not None:
+                context = self.context
+
             arg = self.data_value(self.get_argument(context))
             if arg is None:
                 return []
+            elif arg == '?' and self[0].symbol == '?':
+                raise self.error('XPTY0004', "cannot evaluate a partial function")
 
             try:
                 if isinstance(arg, UntypedAtomic):
@@ -370,7 +305,7 @@ class XPath2Parser(XPath1Parser):
     def schema_constructor(self, atomic_type_name: str, bp: int = 90) \
             -> Type[XPathFunction]:
         """Registers a token class for a schema atomic type constructor function."""
-        if atomic_type_name in {XSD_ANY_ATOMIC_TYPE, XSD_NOTATION}:
+        if atomic_type_name in (XSD_ANY_ATOMIC_TYPE, XSD_NOTATION):
             raise xpath_error('XPST0080')
 
         def nud_(self_: XPathFunction) -> XPathFunction:
@@ -415,15 +350,166 @@ class XPath2Parser(XPath1Parser):
             Type[XPathFunction], ABCMeta(token_class_name, (XPathFunction,), kwargs)
         )
         MutableSequence.register(token_class)
+
+        if self.symbol_table is self.__class__.symbol_table:
+            self.symbol_table = dict(self.__class__.symbol_table)
         self.symbol_table[symbol] = token_class
+        self.tokenizer = None
+
+        return token_class
+
+    def external_function(self,
+                          callback: Callable[..., Any],
+                          name: Optional[str] = None,
+                          prefix: Optional[str] = None,
+                          sequence_types: Tuple[str, ...] = (),
+                          bp: int = 90) -> Type[XPathFunction]:
+        """Registers a token class for an external function."""
+        import inspect
+
+        symbol = name or callback.__name__
+        if not is_ncname(symbol):
+            raise ElementPathValueError(f'{symbol!r} is not a name')
+        elif symbol in self.RESERVED_FUNCTION_NAMES:
+            raise ElementPathValueError(f'{symbol!r} is a reserved function name')
+
+        nargs: NargsType
+        spec = inspect.getfullargspec(callback)
+        if spec.varargs is not None:
+            if spec.args:
+                nargs = len(spec.args), None
+            else:
+                nargs = None
+        elif spec.defaults is None:
+            nargs = len(spec.args)
+        else:
+            nargs = len(spec.args) - len(spec.defaults), len(spec.args)
+
+        if prefix:
+            namespace = self.namespaces[prefix]
+            qname = QName(namespace, f'{prefix}:{symbol}')
+        else:
+            namespace = XPATH_FUNCTIONS_NAMESPACE
+            qname = QName(XPATH_FUNCTIONS_NAMESPACE, f'fn:{symbol}')
+
+        class_name = f'{upper_camel_case(qname.qname)}ExternalFunction'
+        lookup_name = qname.expanded_name
+
+        if self.symbol_table is self.__class__.symbol_table:
+            self.symbol_table = dict(self.__class__.symbol_table)
+
+        if lookup_name in self.symbol_table:
+            msg = f'function {qname.qname!r} is already registered'
+            raise ElementPathValueError(msg)
+        elif symbol not in self.symbol_table or \
+                not issubclass(self.symbol_table[symbol], ProxyToken):
+
+            if symbol in self.symbol_table:
+                token_cls = self.symbol_table[symbol]
+                if not issubclass(token_cls, XPathFunction) \
+                        or token_cls.label == 'kind test':
+                    msg = f'{symbol!r} name collides with {token_cls!r}'
+                    raise ElementPathValueError(msg)
+                if namespace == token_cls.namespace:
+                    msg = f'function {qname.qname!r} is already registered'
+                    raise ElementPathValueError(msg)
+
+                # Copy the token class before register the proxy token
+                self.symbol_table[f'{{{token_cls.namespace}}}{symbol}'] = token_cls
+
+            proxy_class_name = f'{upper_camel_case(qname.local_name)}ProxyToken'
+            kwargs = {
+                'class_name': proxy_class_name,
+                'symbol': symbol,
+                'lbp': bp,
+                'rbp': bp,
+                '__module__': self.__module__,
+                '__qualname__': proxy_class_name,
+                '__return__': None
+            }
+            proxy_class = cast(
+                Type[ProxyToken], ABCMeta(class_name, (ProxyToken,), kwargs)
+            )
+            MutableSequence.register(proxy_class)
+            self.symbol_table[symbol] = proxy_class
+
+        def evaluate_external_function(self_: XPathFunction,
+                                       context: Optional[XPathContext] = None) -> Any:
+            args = []
+            for k in range(len(self_)):
+                arg = self_.get_argument(context, index=k)
+                args.append(arg)
+
+            if sequence_types:
+                for k, (arg, st) in enumerate(zip(args, sequence_types), start=1):
+                    if not match_sequence_type(arg, st, self):
+                        msg_ = f"{ordinal(k)} argument does not match sequence type {st!r}"
+                        raise xpath_error('XPDY0050', msg_)
+
+                result = callback(*args)
+                if not match_sequence_type(result, sequence_types[-1], self):
+                    msg_ = f"Result does not match sequence type {sequence_types[-1]!r}"
+                    raise xpath_error('XPDY0050', msg_)
+                return result
+
+            return callback(*args)
+
+        kwargs = {
+            'class_name': class_name,
+            'symbol': symbol,
+            'namespace': namespace,
+            'label': 'external function',
+            'nargs': nargs,
+            'lbp': bp,
+            'rbp': bp,
+            'evaluate': evaluate_external_function,
+            '__module__': self.__module__,
+            '__qualname__': class_name,
+            '__return__': None
+        }
+        if sequence_types:
+            # Register function signature(s)
+            kwargs['sequence_types'] = sequence_types
+            if self.function_signatures is self.__class__.function_signatures:
+                self.function_signatures = dict(self.__class__.function_signatures)
+
+            if nargs is None:
+                pass  # pragma: no cover
+            elif isinstance(nargs, int):
+                assert len(sequence_types) == nargs + 1
+                self.function_signatures[(qname, nargs)] = 'function({}) as {}'.format(
+                    ', '.join(sequence_types[:-1]), sequence_types[-1]
+                )
+            elif nargs[1] is None:
+                assert len(sequence_types) == nargs[0] + 1
+                self.function_signatures[(qname, nargs[0])] = 'function({}, ...) as {}'.format(
+                    ', '.join(sequence_types[:-1]), sequence_types[-1]
+                )
+            else:
+                assert len(sequence_types) == nargs[1] + 1
+                for arity in range(nargs[0], nargs[1] + 1):
+                    self.function_signatures[(qname, arity)] = 'function({}) as {}'.format(
+                        ', '.join(sequence_types[:arity]), sequence_types[-1]
+                    )
+
+        token_class = cast(
+            Type[XPathFunction], ABCMeta(class_name, (XPathFunction,), kwargs)
+        )
+        MutableSequence.register(token_class)
+
+        self.symbol_table[lookup_name] = token_class
+        self.tokenizer = None
         return token_class
 
     def is_schema_bound(self) -> bool:
-        return 'symbol_table' in self.__dict__
+        return self.schema is not None and 'symbol_table' in self.__dict__
 
     def parse(self, source: str) -> XPathToken:
+        if self.tokenizer is None:
+            self.tokenizer = self.create_tokenizer(self.symbol_table)
+
         root_token = super(XPath1Parser, self).parse(source)
-        if root_token.label == 'sequence type':
+        if root_token.label in ('sequence type', 'function test'):
             raise root_token.error('XPST0003', "not allowed in XPath expression")
 
         if self.schema is None:
@@ -453,7 +539,7 @@ class XPath2Parser(XPath1Parser):
             except KeyError:
                 sequence_type = 'item()*' if isinstance(value, list) else 'item()'
 
-            if not self.match_sequence_type(value, sequence_type):
+            if not match_sequence_type(value, sequence_type, self):
                 message = "Unmatched sequence type for variable {!r}".format(varname)
                 raise xpath_error('XPDY0050', message)
 
